class Boid {

  PVector position = new PVector(random(width), random(height), random(10));
  PVector velocity = new PVector(random(-5, 5), random(-5, 5), random(-5, 5));
  PVector acceleration = new PVector(random(-1, 1), random(-1, 1), random(-1, 1));

  int viewRad = 50;
  float maxForce = 0.2;
  float maxSpeed = 10;

  int viewAngle = 360;
  int currentTargetNum = -1;

  Target currentTarget;

  String mac;
  int ID;

  color f;
  color s;
  color g;
  int shape = 0;

  float size = 10;
  
  float randomOffset = random(255);
  

  Boid(String m, int id) {
    velocity.setMag(random(-.5, .5));
    f = color(random(200)+55);
    g =  color(random(255), random(255), random(255));
    shape = floor(random(3));
    mac = m;
    ID = id;
  }


  PVector align(ArrayList<Boid> boids) {
    PVector avg = new PVector();
    int close = 0;
    for (int i = 0; i < boids.size(); i++) {
      if (inView(boids.get(i))) {
        avg.add(boids.get(i).velocity);
        close++;
      }
    }
    if (close > 0) {
      avg.div(close);
      avg.setMag(maxSpeed);
      avg.sub(velocity);
      avg.limit(maxForce);
    }
    return avg;
  }



  PVector cohesion(ArrayList<Boid> boids) {
    PVector avg = new PVector();
    int close = 0;
    for (int i = 0; i < boids.size(); i++) {
      if (inView(boids.get(i))) {
        avg.add(boids.get(i).position);
        close++;
      }
    }
    if (close > 0) {
      avg.div(close);
      avg.sub(position);
      avg.setMag(maxSpeed);
      avg.sub(velocity);
      avg.limit(maxForce);
    }
    return avg;
  }


  PVector separation(ArrayList<Boid> boids) {
    PVector avg = new PVector();
    int close = 0;
    for (int i = 0; i < boids.size(); i++) {
      if (inView(boids.get(i))) {
        PVector diff = PVector.sub(position, boids.get(i).position);
        float d = dist(boids.get(i).position.x, boids.get(i).position.y, boids.get(i).position.z, position.x, position.y, position.z);
        diff.div(d);
        avg.add(diff);
        close++;
      }
    }
    if (close > 0) {
      avg.div(close);
      avg.setMag(maxSpeed);
      avg.sub(velocity);
      avg.limit(maxForce);
    }
    return avg;
  }

  PVector targetting() {
    if (currentTarget != null) {
      PVector t = currentTarget.position.copy();
      t.sub(position);
      t.setMag(maxSpeed);
      t.sub(velocity);
      //t.limit(maxForce);
      float dist = dist(currentTarget.position.x,currentTarget.position.y,currentTarget.position.z, position.x, position.y, position.z);
      
      t.limit(maxForce);
      
      return t;
    } else {
      return new PVector(0, 0);
    }
  }

  PVector cage() {
    PVector t = currentTarget.position.copy();
    t.mult(0);
    float multiplier = 0;
    PVector loc = currentTarget.position;
    float w = currentTarget.w;
    float h = currentTarget.h;
    float z = currentTarget.z;
    if (position.x < loc.x + w /2 && position.x > loc.x - w /2 && position.y < loc.y + h /2 && position.y > loc.y - h /2 && position.z < loc.z + z /2 && position.z > loc.z - z /2) {
    } else {
      float distX = currentTarget.position.x - position.x;
      float distY = currentTarget.position.y - position.y;
      float distZ = currentTarget.position.z - position.z;

      float fx = map(distX, -w/2, w/2, -1, 1);
      float fy = map(distY, -h/2, h/2, -1, 1);
      float fz = map(distZ, -z/2, z/2, -1, 1);
      multiplier = (abs(fx) + abs(fy) + abs(fz)) * 1.2;

      t.add(new PVector(fx, fy, fz));
    }
    t.setMag(maxSpeed/3);
    t.sub(velocity);
    t.limit(maxForce);
    t.mult(multiplier);
    return t;
  }

  void flock(ArrayList<Boid> boids, float vAl, float vCoh, float vSep, float vTar, float constrain, float speed) {
    maxSpeed = speed;
    acceleration.mult(0);
    PVector alignment = align(boids);
    PVector cohesion = cohesion(boids);
    PVector separation = separation(boids);
    PVector targetting = targetting();
    PVector cage = cage();


    alignment.mult(vAl);
    cohesion.mult(vCoh);
    separation.mult(vSep);
    targetting.mult(vTar);
    cage.mult(constrain);
    

    acceleration.add(alignment);
    acceleration.add(cohesion);
    acceleration.add(separation);
    acceleration.add(targetting);
    acceleration.add(cage);
    
    acceleration.z = acceleration.z/2;
  }

  void pickTarget(ArrayList<Target> targets) {
    currentTargetNum = floor( random( targets.size() ) );
    currentTarget = targets.get( currentTargetNum );
    currentTarget.pick();
  }

  void reTarget(String h) {
    if (currentTargetNum != -1) {
      targets.get(currentTargetNum).unpick();
    }
    for (int i = 0; i < targets.size(); i ++) {
      if (targets.get(i).handle.equals(h)) {
        currentTarget = targets.get(i);
        currentTarget.pick();
        currentTargetNum = i;
      }
    }
  }
  void resetTarget() {
    currentTargetNum = -1;
  }

  void update(ArrayList<Boid> boids) {
    align(boids);
    position.add(velocity);
    velocity.add(acceleration);
    velocity.limit(maxSpeed);
    //edges();
  }
  void adjustSize(float newSize) {
    size = 10 + newSize * 2;
  }
  void show(boolean GODMODE, ArrayList<String> currentGods, boolean DEBUG) {
    pushMatrix();
    float t = size;
    translate(position.x, position.y, position.z);
    //float h = velocity.heading();
    noStroke();
    if(DEBUG) {
      fill(70);
      textSize(8);
      text(ID, 10, 10);
      text(mac, 10, 20);
      //rotate(h + radians(90));
    }
    fill(f);
    sphere(t/2);

    for (int i = 0; i < currentGods.size(); i++) {
      String j = currentGods.get(i);
      if (j.equals(mac)) {
        noFill();
        stroke(255, 0, 255);
        t = size * 2.5;
        rotateZ(radians(frameCount*2 + randomOffset));
        rotateY(radians(frameCount*2 + randomOffset));
        rotateX(radians(frameCount*2 + randomOffset));
        box(t);
        rotateZ(radians(45));
        rotateX(radians(45));
        box(t);
      }
    }
    
    
    popMatrix();
  }

  boolean inView(Boid b) {
    if (b.position != position) {
      float d = dist(b.position.x, b.position.y, position.x, position.y);
      if (d <= viewRad) {
        return true;
        //}
      }
    }
    return false;
  }
  void edges() {
    if (position.x < 0) {
      position.x = width;
    }
    if (position.y < 0) {
      position.y = height;
    }
    if (position.x > width) {
      position.x = 0;
    }
    if (position.y > height) {
      position.y = 0;
    }
  }
}
