
// This application runs in full screen
// Press "d" to adjust the settings


import mqtt.*;
import ch.bildspur.postfx.builder.*;
import ch.bildspur.postfx.pass.*;
import ch.bildspur.postfx.*;


PostFXSupervisor supervisor;
NoisePass noisePass;
BloomPass bloomPass;
SobelPass sobelPass;
PixelatePass pixelPass;
RGBSplitPass rgbSplitPass;
SaturationVibrancePass satPass;

MQTTClient client;

ArrayList<Boid> boids = new ArrayList<Boid>();
ArrayList<Target> targets = new ArrayList<Target>();

boolean connected = false;
boolean DEBUG = false;
boolean VOTING = false;
boolean SHOWQ = false;
boolean RECORDING = false;
boolean STANDALONE = true;

PShape s;

float alignment = 1;
float cohesion = 1;
float separation = 1.5;
float targetting = 1;
float constrain = 1;
float speed = 15;
float lightDist = 0;
float boxDist = -230;

float curSpeed = 0;
float speedTarget = speed;


ArrayList<Slider> sliders = new ArrayList<Slider>();
String[] options = {
  "alignment",
  "cohesion",
  "separation",
  "targetting",
  "constrain",
  "speed",
  "lightDist",
  "boxDist"
};
float[] initials = {
  alignment,
  cohesion,
  separation,
  targetting,
  constrain,
  speed,
  340,
  -230
};

boolean DISPLAY_ENDING = false;
boolean GLITCH = false;
boolean CHAOS = false;
boolean STARDUST = false;
boolean QUESTION_STAY = false;
boolean DISPLAY = true;
boolean GODMODE = false;
ArrayList currentGods = new ArrayList<String>();

String question = "";
void setup() {
  fullScreen(P3D);
  frameRate(30);
  
  for (int i = 0; i < 1; i++) {
    Target t = new Target( width/2, height / 2, "test"+i, "test", boxDist);
    targets.add(t);
  }

  if(STANDALONE) {
    for (int i = 0; i < 150; i++) {
      Boid b = new Boid("TEST_"+i, i);
      b.pickTarget(targets);
      boids.add(b);
    }
  }
  
  supervisor = new PostFXSupervisor(this);
  noisePass = new NoisePass(this, 0.1, 1);
  bloomPass = new BloomPass(this, 0.2, 20, 40);
  sobelPass = new SobelPass(this);
  pixelPass = new PixelatePass(this, 500);
  rgbSplitPass = new RGBSplitPass(this, 100);
    
  
  background(0);
  for (int i = 0; i < options.length; i++) {
    Slider s = new Slider(10, 30 * i + 20, options[i], initials[i], 0, 2);
    if (options[i].equals("speed")) {
      s = new Slider(10, 30 * i + 20, options[i], initials[i], 0, 40);
    }
    if (options[i].equals("lightDist")) {
      s = new Slider(10, 30 * i + 20, options[i], initials[i], -1000, 1000);
    }
    if (options[i].equals("boxDist")) {
      s = new Slider(10, 30 * i + 20, options[i], initials[i], -1000, 1000);
    }
    sliders.add(s);
  }
}

void draw() {

  if (GODMODE) {
    background(0);
  } else {
    background(0);
  }
  lights();


  pushMatrix();
  noFill();
  translate(width/2 - width/4, height/2, lightDist);
  pointLight(50, 50, 50, 0, 0, 0);
  //sphere(20);
  translate(width/4, 0, 0);
  pointLight(40, 40, 40, 0, 0, 0);
  //sphere(20);
  translate(width/4, 0, 0);
  pointLight(50, 50, 50, 0, 0, 0);
  //sphere(20);
  popMatrix();


  if (!connected && !STANDALONE) {
    try {
      client = new MQTTClient(this);
      client.connect("mqtt://10.10.0.2:1883", "processingViz");
      client.subscribe("controllers/+/move");
      client.subscribe("controllers/+/im_alive");
      client.subscribe("services/director/+");
      client.subscribe("services/analyst/godmode");
      connected = true;
    }
    catch(RuntimeException e) {
      println(e);
      delay(1000);
    }
  } else {

    if (VOTING || CHAOS) {
      speedTarget = speed * 2;
    } else {
      speedTarget = speed;
    }
    curSpeed += (speedTarget - curSpeed) / 10;
    for (Boid b : boids) {
      b.flock(boids, alignment, cohesion, separation, targetting, constrain, curSpeed);
    }

    for (Boid b : boids) {
      b.update(boids);
      b.show(GODMODE, currentGods, DEBUG);
    }

    if (!question.contains("PAD") && SHOWQ) {
      fill(255);
      textAlign(CENTER);
      textSize(24);
      text(question, width / 2, 100);
    }
  }



    
    
  if (CHAOS) {
    supervisor.render();
      //supervisor.pass(pixelPass);
      supervisor.pass(noisePass);
      supervisor.pass(bloomPass);
      //supervisor.pass(sobelPass);
      supervisor.pass(rgbSplitPass);
      supervisor.compose();
  } else {
    
    supervisor.render();
      supervisor.pass(noisePass);
      supervisor.pass(bloomPass);
      supervisor.compose();
  }
  for (Target t : targets) {
    t.update(boxDist);
    t.show(SHOWQ, DEBUG);
  }
  if (DEBUG) {
    for (int i = 0; i < sliders.size(); i++) {
      sliders.get(i).show();
    }
    textSize(10);
    fill(255);
    textAlign(LEFT);
    text(frameRate, 300, 30);
    text("NUM: "+boids.size(), 300, 45);
    text("CHAOS: "+CHAOS, 300, 75);
    text("GLITCH: "+GLITCH, 300, 90);
    text("DISPLAY: "+DISPLAY, 300, 105);
    text("STARDUST: "+STARDUST, 300, 120);
    text("RECORDING: "+RECORDING, 300, 135);
  }
}

void mousePressed() {
  for (Slider s : sliders) {
    s.downHandler();
  }
}
void mouseReleased() {
  for (Slider s : sliders) {
    s.upHandler();
  }
}
void mouseDragged() {
  for (int i = 0; i < sliders.size(); i++) {
    sliders.get(i).dragHandler();
    updateSetting(sliders.get(i).getValue(), options[i]);
  }
}


void updateSetting(float val, String setting) {
  switch(setting) {
  case "alignment":
    alignment = val;
    break;
  case "cohesion":
    cohesion = val;
    break;
  case "separation":
    separation = val;
    break;
  case "targetting":
    targetting = val;
    break;
  case "constrain":
    constrain = val;
    break;
  case "speed":
    speed = val;
  case "lightDist":
    lightDist = val;
  case "boxDist":
    boxDist = val;
    break;
  }
}



void keyPressed() {
  if (key == 'k') {
    if (connected) {
      int p = floor(random(targets.size()));
      int id = floor(random(boids.size()));
      //client.publish("choose", "{\"picked\":"+p+", \"id\":"+id+"}");
    }
  }
  if (key == 'l') {
    if (connected) {
      client.publish("services/tempViz/cue", "{\"action\": \"request\"}");
    }
  }

  if (key == 'b') {
    for (Target t : targets) {
      t.reset();
    }
    for (Boid b : boids) {
      b.pickTarget(targets);
    }
  }
  if (key == 'd') {
    DEBUG = !DEBUG;
    if (DEBUG) {
      cursor(HAND);
    } else {
      noCursor();
    }
  }
}

void messageReceived(String topic, byte[] payload) {

  if (topic.contains("controller")) {
    String mac = topic.split("/")[1];
    boolean found = false;
    for (int i = 0; i < boids.size(); i++) {
      if (boids.get(i).mac.equals(mac)) {
        found = true;
      }
    }
    if (!found) {
      Boid b = new Boid(mac, boids.size());
      b.pickTarget(targets);
      boids.add(b);
    }


    if (topic.contains("move")) {
      // we should move one
      String message = new String(payload);
      //println(message);
      JSONObject m = parseJSONObject(message);
      String vote = m.getString("handle");
      float size = m.getFloat("size");
      for (int i = 0; i < boids.size(); i++) {
        if (boids.get(i).mac.equals(mac)) {
          boids.get(i).adjustSize(size);
          boids.get(i).reTarget(vote);
        }
      }
    }
  } else if (topic.contains("cue")) {
    String message = new String(payload);
    //println(message);
    JSONArray m = parseJSONArray(message);
    targets.clear();
    for (int i = 0; i < m.size(); i++) {
      JSONObject obj = m.getJSONObject(i);
      question = obj.getString("question");
      JSONArray o = obj.getJSONArray("options");
      for (int j = 0; j < o.size(); j++) {
        JSONObject opt = o.getJSONObject(j);
        if (m.size() == 1) {
          Target t = new Target( width/2  - (width/4) + ((width/ 2) * j), height / 2, opt.getString("optionHandle"), opt.getString("text"), boxDist);
          targets.add(t);
        } else {
          Target t = new Target( width/2  - (width/4) - (width/8) + ((width/ 4) * j) + (width / 2 * (m.size() - 1 - i)), height / 2, opt.getString("optionHandle"), opt.getString("text"), boxDist);
          targets.add(t);
        }
      }
    }
    for (int i = 0; i < boids.size(); i++) {
      boids.get(i).resetTarget();
    }
  } else if (!topic.contains("im_alive")) {
    String message = new String(payload);
    //println(message);
    JSONObject m = parseJSONObject(message);

    if (topic.contains("vote")) {
      String action = m.getString("action");
      if (action.equals("enable")) {
        VOTING = true;
        SHOWQ = true;
      } else if (action.equals("disable")) {
        VOTING = false;
        if (!QUESTION_STAY) {
          SHOWQ = false;
        }
      }
    }
    if (topic.equals("services/director/godmode")) {
      String action = m.getString("action");
      if (action.equals("enable")) {
        println("GODMODE ENABLED");
        GODMODE = true;
      } else if (action.equals("disable")) {
        GODMODE = false;
        currentGods.clear();
      }
    }
    if (topic.equals("services/analyst/godmode")) {
      if (m.getString("action").equals("info")) {
        currentGods.clear();
        try {
          JSONArray selected = m.getJSONArray("controllers");
          for (int i = 0; i < selected.size(); i++) {
            JSONObject j = selected.getJSONObject(i);
            if (j.getInt("god") == 1) {
              String h = j.getString("id");
              currentGods.add(h);
            }
          }
        }
        catch(RuntimeException e) {
          println(e);
          println(m);
        }
      }
    }
    if (topic.contains("chaos")) {
      if (m.getString("action").equals("enable")) {
        CHAOS = true;
        speed = 10;
      } else {
        CHAOS = false;
        speed = 4;
      }
    }
    if (topic.contains("display")) {
      if (m.getString("action").equals("enable")) {
        DISPLAY = true;
      } else {
        DISPLAY = false;
      }
    }
    if (topic.contains("stardust")) {
      if (m.getString("action").equals("enable")) {
        STARDUST = true;
        speed = 0.01;
        //for(Boid b : boids) {
        //  b.size = 1;
        //}
      } else {
        STARDUST = false;
        speed = 4;
      }
    }
    if (topic.contains("glitch")) {
      if (m.getString("action").equals("enable")) {
        GLITCH = true;
      } else {
        GLITCH = false;
      }
    }
    if (topic.contains("questionStay")) {
      if (m.getString("action").equals("enable")) {
        QUESTION_STAY = true;
      } else {
        QUESTION_STAY = false;
      }
    }
    if (topic.contains("displayEnding")) {
      if (m.getString("action").equals("enable")) {
        DISPLAY_ENDING = true;
      } else {
        DISPLAY_ENDING = false;
      }
    }
  }
}
